"""Script to assess the performance of the SQLite backend."""

import shutil
import tempfile
import time
import random
import uuid
import numpy as np
from contextlib import contextmanager
from pathlib import Path

from sifts import Collection


@contextmanager
def timer(title: str):
    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        elapsed_time = end_time - start_time
        if elapsed_time < 1e-9:
            print(f"{title:<25} Took < 1 ns")
        elif elapsed_time < 1e-6:
            print(f"{title:<25} Took {elapsed_time / 1e-9:.0f} ns")
        elif elapsed_time < 1e-3:
            print(f"{title:<25} Took {elapsed_time / 1e-6:.0f} µs")
        elif elapsed_time < 0.5:
            print(f"{title:<25} Took {elapsed_time / 1e-3:.0f} ms")
        else:
            print(f"{title:<25} Took {elapsed_time:.2f} s")


word_list = [
    "Lorem",
    "ipsum",
    "dolor",
    "sit",
    "amet",
    "consectetur",
    "adipiscing",
    "elit",
    "Sed",
    "do",
    "eiusmod",
    "tempor",
    "incididunt",
    "ut",
    "labore",
    "et",
    "dolore",
    "magna",
    "aliqua",
    "Ut",
    "enim",
    "ad",
    "minim",
    "veniam",
    "quis",
    "nostrud",
    "exercitation",
    "ullamco",
    "laboris",
    "nisi",
    "ut",
    "aliquip",
    "ex",
    "ea",
    "commodo",
    "consequat",
    "Duis",
    "aute",
    "irure",
    "dolor",
    "in",
    "reprehenderit",
    "in",
    "voluptate",
    "velit",
    "esse",
    "cillum",
    "dolore",
    "eu",
    "fugiat",
    "nulla",
    "pariatur",
    "Excepteur",
    "sint",
    "occaecat",
    "cupidatat",
    "non",
    "proident",
    "sunt",
    "in",
    "culpa",
    "qui",
    "officia",
    "deserunt",
    "mollit",
    "anim",
    "id",
    "est",
    "laborum",
]


def get_document(n: int = 20) -> str:
    """Generate a string of n words."""
    return " ".join(random.choices(word_list, k=n))


def get_word() -> str:
    """Get a random word."""
    return random.choice(word_list)


def run_timing():
    """Run timing."""
    tmp_dir = Path(tempfile.mkdtemp())
    path = tmp_dir / "search_engine.db"

    with timer("Create database table"):
        Collection(f"sqlite:///{path}", name="123")

    with timer("Instantiate again"):
        engine = Collection(f"sqlite:///{path}", name="123")

    print("-- Full-text search --")
    run_add_update_delete(engine)

    def f(documents):
        return [np.random.rand(384) for _ in documents]

    engine = Collection(
        f"sqlite:///{path}", name="456", embedding_function=f, use_fts=False
    )

    print("-- Vector search --")
    run_add_update_delete(engine)

    engine = Collection(f"sqlite:///{path}", name="789", embedding_function=f)

    print("-- Both --")
    run_add_update_delete(engine)

    shutil.rmtree(tmp_dir)


def run_add_update_delete(engine, n=100000):
    """Run some long running operations."""

    N = n
    random.seed(711)

    with timer("Count on empty"):
        engine.count()

    with timer("Create random documents"):
        contents = [get_document() for _ in range(N)]

    with timer("Create random IDs"):
        ids = [str(uuid.uuid4()) for _ in range(N)]

    with timer("Create random metadata"):
        metadatas = [{"k1": get_word(), "k2": get_word()} for _ in range(N)]

    if engine.embedding_function:
        with timer("Embeddings"):
            vectors = engine.embedding_function(contents)
        with timer("Format vectors"):
            engine._format_vectors(vectors)

    with timer("Add documents"):
        engine.add(contents, ids, metadatas)

    with timer("Retrieve metadata"):
        engine.get()

    with timer("Get all"):
        engine.get()

    with timer("Count on full"):
        engine.count()

    with timer("Delete documents"):
        engine.delete_all()

    with timer("Add documents"):
        engine.add(contents, ids, metadatas)

    with timer("Update documents"):
        engine.add(contents, ids, metadatas)

    with timer("Query"):
        engine.query("reprehenderit", vector_search=bool(engine.embedding_function))


if __name__ == "__main__":
    run_timing()



"""Unit tests for Postgres backend."""

import time

import pytest
import psycopg2
from sifts.core import CollectionPostgreSQL
from psycopg2 import OperationalError

TEST_DB_DSN = "postgresql://testuser:testpass@localhost:5432/testdb"


@pytest.fixture(scope="session")
def docker_compose_file(pytestconfig):
    return "docker-compose.yml"


@pytest.fixture(scope="session")
def postgres_service(docker_services):
    """Wait for the PostgreSQL service to be up and running."""

    def is_postgres_healthy():
        try:
            conn = psycopg2.connect(
                dbname="testdb",
                user="testuser",
                password="testpass",
                host="localhost",
                port=5432,
            )
            conn.close()
            return True
        except OperationalError:
            return False

    timeout = 30
    pause = 0.5
    start_time = time.time()

    while time.time() - start_time < timeout:
        if is_postgres_healthy():
            return
        time.sleep(pause)

    raise TimeoutError("PostgreSQL service did not become healthy in time")


@pytest.fixture
def search_engine():
    engine = CollectionPostgreSQL(dsn=TEST_DB_DSN, name="my_name")
    yield engine
    with engine.conn() as conn:
        # conn.execute("DROP TRIGGER tsvectorupdate ON documents CASCADE;")
        conn.execute("TRUNCATE TABLE documents RESTART IDENTITY CASCADE;")


def test_add_document(postgres_service, search_engine):
    content = ["test content"]
    assert search_engine.count() == 0
    ids = search_engine.add(content)
    assert len(ids) == 1
    assert search_engine.count() == 1


def test_query_document(postgres_service, search_engine):
    content = ["test query content"]
    search_engine.add(content)
    results = search_engine.query("content")
    assert results["total"] == 1
    assert len(results["results"]) == 1


def test_update_document(postgres_service, search_engine):
    content = ["initial content"]
    ids = search_engine.add(content)
    results = search_engine.query("initial")
    assert results["total"] == 1
    assert len(results["results"]) == 1
    updated_content = ["updated content"]
    search_engine.update(ids, updated_content)
    results = search_engine.query("updated")
    assert results["total"] == 1
    assert len(results["results"]) == 1
    results = search_engine.query("content")
    assert results["total"] == 1
    assert len(results["results"]) == 1
    results = search_engine.query("initial")
    assert results["total"] == 0
    assert len(results["results"]) == 0


def test_delete_document(postgres_service, search_engine):
    content = ["content to delete"]
    assert search_engine.count() == 0
    ids = search_engine.add(content)
    assert search_engine.count() == 1
    search_engine.delete(ids)
    assert search_engine.count() == 0
    results = search_engine.query("delete")
    assert results["total"] == 0
    assert len(results["results"]) == 0


def test_add(postgres_service, search_engine):
    assert search_engine.query("Lorem") == {"total": 0, "results": []}
    ids1 = search_engine.add(["Lorem ipsum dolor"])
    ids2 = search_engine.add(["sit amet"])
    assert len(search_engine.query("Lorem")["results"]) == 1
    assert search_engine.query("Lorem")["results"][0]["id"] == ids1[0]


def test_query_wildcard(postgres_service, search_engine):
    assert search_engine.query("Lorem") == {"total": 0, "results": []}
    ids1 = search_engine.add(["Lorem ipsum dolor"])
    ids2 = search_engine.add(["sit amet"])
    assert len(search_engine.query("am*")["results"]) == 1
    assert search_engine.query("am*")["results"][0]["id"] == ids2[0]
    assert len(search_engine.query("ame*")["results"]) == 1


def test_query_pr(postgres_service, search_engine):
    assert search_engine.query("Lorem") == {"total": 0, "results": []}
    ids1 = search_engine.add(["Lorem ipsum dolor"])
    ids2 = search_engine.add(["sit amet"])
    assert len(search_engine.query("Lorem or amet")["results"]) == 2


def test_add_name(postgres_service, search_engine):
    search_engine_2 = CollectionPostgreSQL(dsn=TEST_DB_DSN, name="my_other_name")
    assert search_engine_2.query("Lorem") == {"total": 0, "results": []}
    search_engine_2.add(["Lorem ipsum dolor"])
    assert len(search_engine_2.query("Lorem")["results"]) == 1
    assert len(search_engine.query("Lorem")["results"]) == 0


def test_add_id(postgres_service, search_engine):
    ids = search_engine.add(["x"])
    assert len(ids) == 1
    assert len(ids[0]) == 36  # is UUIDv4
    ids = search_engine.add(["y"], ids=["my_id"])
    assert ids == ["my_id"]
    res = search_engine.query("y")
    assert len(res["results"]) == 1
    res = res["results"]
    assert res[0]["id"] == "my_id"
    # does not raise, but updates
    search_engine.add(["z"], ids=["my_id"])
    res = search_engine.query("y")
    assert len(res["results"]) == 0
    res = search_engine.query("z")
    assert len(res["results"]) == 1


def test_update(postgres_service, search_engine):
    ids = search_engine.add(["Lorem ipsum"])
    res = search_engine.query("Lorem")
    assert len(res["results"]) == 1
    res = res["results"]
    assert res[0]["id"] == ids[0]
    search_engine.update(ids=ids, contents=["dolor sit"])
    res = search_engine.query("Lorem")
    assert len(res["results"]) == 0
    res = search_engine.query("sit")
    assert len(res["results"]) == 1
    res = res["results"]
    assert res[0]["id"] == ids[0]


def test_delete(postgres_service, search_engine):
    ids = search_engine.add(["Lorem ipsum"])
    res = search_engine.query("Lorem")
    assert len(res["results"]) == 1
    search_engine.delete(ids)
    res = search_engine.query("Lorem")
    assert len(res["results"]) == 0
    search_engine.delete(ids)


def test_query_multiple(postgres_service, search_engine):
    search_engine.add(["Lorem ipsum dolor"])
    search_engine.add(["sit amet"])
    assert len(search_engine.query("Lorem ipsum")["results"]) == 1
    assert len(search_engine.query("sit amet")["results"]) == 1
    assert len(search_engine.query("Lorem sit")["results"]) == 0


def test_query_order(postgres_service, search_engine):
    search = search_engine
    search.add(["Lorem"], metadatas=[{"k1": "a", "k2": "c"}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": "b", "k2": "c"}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": "c", "k2": "c"}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": "d", "k2": "b"}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": "e", "k2": "b"}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": "f", "k2": "b"}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": "g", "k2": "a"}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": "h", "k2": "a"}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": "i", "k2": "a"}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    res = search.query("Lorem")
    assert len(res["results"]) == 10
    # k1
    res = search.query("Lorem", order_by="k1")
    res = res["results"]
    assert len(res) == 10
    assert [r["id"][1:] for r in res] == list("1234567890")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("abcdefghi0")
    # +k1
    res = search.query("Lorem", order_by="+k1")["results"]
    assert [r["id"][1:] for r in res] == list("1234567890")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("abcdefghi0")
    # -k1
    res = search.query("Lorem", order_by="-k1")["results"]
    assert [r["id"][1:] for r in res] == list("0987654321")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("0ihgfedcba")
    # k2,k1
    res = search.query("Lorem", order_by=["k2", "k1"])["results"]
    assert [r["id"][1:] for r in res] == list("7894561230")
    assert [(r["metadata"] or {}).get("k2", "0") for r in res] == list("aaabbbccc0")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("ghidefabc0")
    # k2,-k1
    res = search.query("Lorem", order_by=["k2", "-k1"])["results"]
    assert [r["id"][1:] for r in res] == list("9876543210")
    assert [(r["metadata"] or {}).get("k2", "0") for r in res] == list("aaabbbccc0")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("ihgfedcba0")


def test_query_limit_offset(postgres_service, search_engine):
    search = search_engine
    search.add(["Lorem"], metadatas=[{"k1": "a", "k2": "c"}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": "b", "k2": "c"}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": "c", "k2": "c"}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": "d", "k2": "b"}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": "e", "k2": "b"}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": "f", "k2": "b"}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": "g", "k2": "a"}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": "h", "k2": "a"}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": "i", "k2": "a"}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    res = search.query("Lorem", order_by="k1")
    assert len(res["results"]) == 10
    res = search.query("Lorem", order_by="k1", limit=0)
    assert len(res["results"]) == 10
    res = search.query("Lorem", order_by="k1", limit=3)
    assert len(res["results"]) == 3
    res = res["results"]
    assert [r["id"][1:] for r in res] == list("123")
    res = search.query("Lorem", order_by="k1", limit=3, offset=3)
    assert len(res["results"]) == 3
    res = res["results"]
    assert [r["id"][1:] for r in res] == list("456")
    res = search.query("Lorem", order_by="k1", limit=3, offset=8)
    assert len(res["results"]) == 2
    res = res["results"]
    assert [r["id"][1:] for r in res] == list("90")


def test_query_where(postgres_service, search_engine):
    search = search_engine
    search.add(["Lorem"], metadatas=[{"k1": "a", "k2": "c"}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": "b", "k2": "c"}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": "c", "k2": "c"}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": "d", "k2": "b"}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": "e", "k2": "b"}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": "f", "k2": "b"}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": "g", "k2": "a"}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": "h", "k2": "a"}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": "i", "k2": "a"}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    res = search.query("Lorem", where={"k2": "a"}, order_by="k1")
    assert len(res["results"]) == 3
    res = search.query("Lorem", where={"k2": {"$eq": "a"}}, order_by="k1")
    assert res["total"] == 3
    res = res["results"]
    assert len(res) == 3
    res = search.query("Lorem", where={"k2": {"$gt": "a"}}, order_by="k1")
    assert res["total"] == 6
    res = res["results"]
    assert len(res) == 6
    res = search.query("Lorem", where={"k2": {"$lt": "a"}}, order_by="k1")
    assert res["total"] == 0
    res = res["results"]
    assert len(res) == 0


def test_query_where_num(postgres_service, search_engine):
    search = search_engine
    search.add(["Lorem"], metadatas=[{"k1": 1, "k2": 3}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": 2, "k2": 3}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": 3, "k2": 3}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": 4, "k2": 2}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": 5, "k2": 2}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": 6, "k2": 2}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": 7, "k2": 1}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": 8, "k2": 1}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": 9, "k2": 1}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    res = search.query("Lorem", where={"k2": 1}, order_by="k1")
    assert res["total"] == 3
    res = res["results"]
    assert len(res) == 3
    res = search.query("Lorem", where={"k2": {"$eq": 1}}, order_by="k1")
    assert res["total"] == 3
    res = res["results"]
    assert len(res) == 3
    res = search.query("Lorem", where={"k2": {"$gt": 1}}, order_by="k1")
    assert res["total"] == 6
    res = res["results"]
    assert len(res) == 6
    res = search.query("Lorem", where={"k2": {"$lt": 1}}, order_by="k1")
    assert res["total"] == 0
    res = res["results"]
    assert len(res) == 0


def test_query_where_in(postgres_service, search_engine):
    search = search_engine
    search.add(["Lorem"], metadatas=[{"k1": "a", "k2": "c"}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": "b", "k2": "c"}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": "c", "k2": "c"}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": "d", "k2": "b"}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": "e", "k2": "b"}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": "f", "k2": "b"}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": "g", "k2": "a"}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": "h", "k2": "a"}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": "i", "k2": "a"}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    with pytest.raises(ValueError):
        # wrong operator
        search.query("Lorem", where={"k1": {"in": "a"}})
    res = search.query(
        "Lorem", where={"k1": {"$in": ["a", "b", "c", "d"]}}, order_by="k1"
    )
    assert res["total"] == 4
    res = res["results"]
    assert len(res) == 4
    assert [r["id"][1:] for r in res] == list("1234")
    res = search.query(
        "Lorem", where={"k1": {"$nin": ["a", "b", "c", "d"]}}, order_by="k1"
    )
    assert res["total"] == 5
    res = res["results"]
    assert len(res) == 5
    assert [r["id"][1:] for r in res] == list("56789")


def test_all_docs(postgres_service, search_engine):
    search = search_engine
    search.add(["Lorem ipsum dolor"])
    search.add(["sit amet"])
    res = search.get()
    assert len(res["results"]) == 2
    assert res["total"] == 2


def test_instantiate_twice(postgres_service, search_engine):
    search = search_engine
    search.add(["Lorem ipsum dolor"])
    res = search.get()
    assert len(res["results"]) == 1
    search = CollectionPostgreSQL(dsn=TEST_DB_DSN, name="my_name")
    res = search.get()
    assert len(res["results"]) == 1


def test_vector(postgres_service, search_engine):
    vectors = {"Lorem ipsum dolor": [0.0, 0.0, 0.0], "sit amet": [0, 0.5, 0]}

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionPostgreSQL(dsn=TEST_DB_DSN, name="vector", embedding_function=f)
    search.add(["Lorem ipsum dolor", "sit amet"])
    with search.conn() as conn:
        conn.execute("SELECT embedding FROM documents")
        vectors = conn.fetchall()
        vectors = [v[0] for v in vectors]
        assert len(vectors) == 2
        assert isinstance(vectors[0], str)
        assert vectors[0] == "[0,0,0]"
        assert vectors[1] == "[0,0.5,0]"


def test_vector_query(postgres_service, search_engine):
    vectors = {
        "Lorem ipsum dolor": [1, 1, 1],
        "sit amet": [1, -1, 1],
        "consectetur": [-1, -1, 1],
        "adipiscing": [-1, -1, -1],
    }

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionPostgreSQL(dsn=TEST_DB_DSN, name="vector", embedding_function=f)
    search.add(["Lorem ipsum dolor", "sit amet"])
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    # limit & offset
    res = search.query("consectetur", vector_search=True, offset=0, limit=1)
    assert res["total"] == 2
    assert len(res["results"]) == 1
    assert res["results"][0]["content"] == "sit amet"
    res = search.query("consectetur", vector_search=True, offset=1, limit=1)
    assert res["total"] == 2
    assert len(res["results"]) == 1
    assert res["results"][0]["content"] == "Lorem ipsum dolor"
    res = search.query("consectetur", vector_search=True, offset=2)
    assert res["total"] == 0
    assert len(res["results"]) == 0


def test_vector_query_fts(postgres_service, search_engine):
    vectors = {
        "Lorem ipsum dolor": [1, 1, 1],
        "sit amet": [1, -1, 1],
        "consectetur": [-1, -1, 1],
        "adipiscing": [-1, -1, -1],
    }

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionPostgreSQL(dsn=TEST_DB_DSN, name="vector", embedding_function=f)
    search.add(["Lorem ipsum dolor", "sit amet"])
    res = search.query("Lorem", vector_search=False)
    assert res["total"] == 1
    assert res["results"][0]["content"] == "Lorem ipsum dolor"


def test_vector_update(postgres_service, search_engine):
    vectors = {
        "Lorem ipsum dolor": [1, 1, 1],
        "sit amet": [1, -1, 1],
        "consectetur": [-1, -1, 1],
        "adipiscing": [-1, -1, -1],
    }

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionPostgreSQL(dsn=TEST_DB_DSN, name="vector", embedding_function=f)
    ids = search.add(["Lorem ipsum dolor", "sit amet"])
    res = search.query("Lorem", vector_search=False)
    assert res["total"] == 1
    res = search.query("consectetur", vector_search=False)
    assert res["total"] == 0
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][0]["id"] == ids[1]
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    assert res["results"][1]["id"] == ids[0]
    # update: switch order
    search.update(ids=ids, contents=["sit amet", "Lorem ipsum dolor"])
    res = search.query("Lorem", vector_search=False)
    assert res["total"] == 1
    res = search.query("consectetur", vector_search=False)
    assert res["total"] == 0
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][0]["id"] == ids[0]
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    assert res["results"][1]["id"] == ids[1]


def test_vector_update_nofts(postgres_service, search_engine):
    vectors = {
        "Lorem ipsum dolor": [1, 1, 1],
        "sit amet": [1, -1, 1],
        "consectetur": [-1, -1, 1],
        "adipiscing": [-1, -1, -1],
    }

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionPostgreSQL(
        dsn=TEST_DB_DSN, name="vector", embedding_function=f, use_fts=False
    )
    ids = search.add(["Lorem ipsum dolor", "sit amet"])
    with pytest.raises(ValueError):
        res = search.query("Lorem", vector_search=False)
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][0]["id"] == ids[1]
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    assert res["results"][1]["id"] == ids[0]
    # update: switch order
    search.update(ids=ids, contents=["sit amet", "Lorem ipsum dolor"])
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][0]["id"] == ids[0]
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    assert res["results"][1]["id"] == ids[1]


def test_query_hyphenated_word(postgres_service, search_engine):
    search_engine.add(["This is a test-word example"])
    search_engine.add(["Another document without it"])
    res = search_engine.query("test-word")
    assert res["total"] == 1
    assert len(res["results"]) == 1
    assert res["results"][0]["content"] == "This is a test-word example"


def test_query_hyphenated_wildcard(postgres_service, search_engine):
    search_engine.add(["This is a test-word example"])
    search_engine.add(["This is a test-word-extended version"])
    search_engine.add(["Another document without it"])
    res = search_engine.query("test-word*")
    assert res["total"] == 2
    assert len(res["results"]) == 2


def test_query_wildcard_end(postgres_service, search_engine):
    search_engine.add(["testing document"])
    search_engine.add(["tester document"])
    search_engine.add(["tests document"])
    search_engine.add(["another document"])
    res = search_engine.query("test*")
    assert res["total"] == 3
    assert len(res["results"]) == 3


def test_query_wildcard_middle(postgres_service, search_engine):
    # Note: PostgreSQL tsquery only supports trailing wildcards (prefix matching)
    # Wildcards in the middle or at the beginning are not supported
    search_engine.add(["testing document"])
    search_engine.add(["tester document"])
    search_engine.add(["another document"])
    # Use trailing wildcard instead - this will match both "testing" and "tester"
    res = search_engine.query("test*")
    assert res["total"] == 2


def test_query_wildcard_beginning(postgres_service, search_engine):
    # Note: PostgreSQL tsquery does not support leading wildcards
    search_engine.add(["testing document"])
    search_engine.add(["interesting document"])
    search_engine.add(["another document"])
    res = search_engine.query("*sting")
    # This won't match anything because leading wildcards aren't supported
    assert res["total"] == 0


def test_query_apostrophe(postgres_service, search_engine):
    search_engine.add(["It's a test document"])
    search_engine.add(["Another document"])
    res = search_engine.query("it's")
    assert res["total"] == 1
    assert res["results"][0]["content"] == "It's a test document"



from sifts.core import QueryParser


def test_trim_sqlite():
    query = " Lorem\t"
    assert str(QueryParser(query)) == "Lorem"


def test_and_sqlite():
    query = "Lorem and ipsum"
    assert str(QueryParser(query)) == "Lorem AND ipsum"


def test_or_sqlite():
    query = "Lorem or ipsum"
    assert str(QueryParser(query)) == "Lorem OR ipsum"


def test_wildcard_sqlite():
    query = "Lor*"
    assert str(QueryParser(query)) == "Lor*"


def test_wildcard_and_sqlite():
    query = "Lor* and ips*"
    assert str(QueryParser(query)) == "Lor* AND ips*"


def test_trim_postgres():
    query = " Lorem\t"
    assert str(QueryParser(query, backend="postgresql")) == "Lorem"


def test_and_postgres():
    query = "Lorem and ipsum"
    assert str(QueryParser(query, backend="postgresql")) == "Lorem & ipsum"


def test_and_postgres_upper():
    query = "Lorem AND ipsum"
    assert str(QueryParser(query, backend="postgresql")) == "Lorem & ipsum"


def test_or_postgres():
    query = "Lorem or ipsum"
    assert str(QueryParser(query, backend="postgresql")) == "Lorem | ipsum"


def test_wildcard_postgres():
    query = "Lor*"
    assert str(QueryParser(query, backend="postgresql")) == "Lor:*"


def test_wildcard_and_postgres():
    query = "Lor* and ips*"
    assert str(QueryParser(query, backend="postgresql")) == "Lor:* & ips:*"


def test_hyphen_sqlite():
    query = "test-word"
    assert str(QueryParser(query)) == '"test-word"'


def test_hyphen_postgres():
    query = "test-word"
    assert str(QueryParser(query, backend="postgresql")) == '"test-word"'


def test_hyphen_multiple_sqlite():
    query = "test-word and another-word"
    assert str(QueryParser(query)) == '"test-word" AND "another-word"'


def test_hyphen_multiple_postgres():
    query = "test-word and another-word"
    assert str(QueryParser(query, backend="postgresql")) == '"test-word" & "another-word"'


def test_hyphen_wildcard_end_sqlite():
    query = "test-word*"
    assert str(QueryParser(query)) == '"test-word"*'


def test_hyphen_wildcard_end_postgres():
    query = "test-word*"
    assert str(QueryParser(query, backend="postgresql")) == '"test-word":*'


def test_hyphen_wildcard_middle_sqlite():
    # Mid-word wildcards are not supported, left as-is
    query = "test-wo*rd"
    assert str(QueryParser(query)) == '"test-wo"*rd'


def test_hyphen_wildcard_middle_postgres():
    # Mid-word wildcards are not supported, left as-is
    query = "test-wo*rd"
    assert str(QueryParser(query, backend="postgresql")) == '"test-wo"*rd'


def test_hyphen_wildcard_beginning_sqlite():
    # Leading wildcards are stripped, then hyphenated word is quoted
    query = "*test-word"
    assert str(QueryParser(query)) == '"test-word"'


def test_hyphen_wildcard_beginning_postgres():
    # Leading wildcards are stripped, then hyphenated word is quoted
    query = "*test-word"
    assert str(QueryParser(query, backend="postgresql")) == '"test-word"'


def test_wildcard_end_only_sqlite():
    query = "testword*"
    assert str(QueryParser(query)) == "testword*"


def test_wildcard_end_only_postgres():
    query = "testword*"
    assert str(QueryParser(query, backend="postgresql")) == "testword:*"


def test_wildcard_middle_only_sqlite():
    # Mid-word wildcards are not supported by SQLite FTS5, left as-is
    query = "test*word"
    assert str(QueryParser(query)) == "test*word"


def test_wildcard_middle_only_postgres():
    # Mid-word wildcards are not supported by PostgreSQL tsquery, left as-is
    query = "test*word"
    assert str(QueryParser(query, backend="postgresql")) == "test*word"


def test_wildcard_beginning_only_sqlite():
    # Leading wildcards are not supported by SQLite FTS5, stripped out
    query = "*testword"
    assert str(QueryParser(query)) == "testword"


def test_wildcard_beginning_only_postgres():
    # Leading wildcards are not supported by PostgreSQL tsquery, stripped out
    query = "*testword"
    assert str(QueryParser(query, backend="postgresql")) == "testword"


def test_apostrophe_sqlite():
    query = "it's"
    assert str(QueryParser(query)) == '"it\'s"'


def test_apostrophe_postgres():
    query = "it's"
    assert str(QueryParser(query, backend="postgresql")) == '"it\'s"'



"""Unit tests."""

import os
import sqlite3

import numpy as np
import pytest
from sifts.core import CollectionSQLite


def test_init(tmp_path):
    path = tmp_path / "search_engine.db"
    CollectionSQLite(path, name="123")
    assert os.path.isfile(path)
    conn = sqlite3.connect(path)
    cursor = conn.cursor()
    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='documents';"
    )
    assert cursor.fetchone() is not None
    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='documents_fts';"
    )
    assert cursor.fetchone() is not None


def test_collection_names(tmp_path):
    path = tmp_path / "search_engine.db"
    with pytest.raises(ValueError):
        CollectionSQLite(path, name="1 2")
    with pytest.raises(ValueError):
        CollectionSQLite(path, name=" abc")
    CollectionSQLite(path, name="1+2")
    CollectionSQLite(path, name="1-2")
    CollectionSQLite(path, name="ab/c")
    CollectionSQLite(path, name="abc")


def test_add(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    assert search.query("Lorem") == {"results": [], "total": 0}
    # assert search.count() == 0
    search.count()
    ids1 = search.add(["Lorem ipsum dolor"])
    ids2 = search.add(["sit amet"])
    assert len(search.query("Lorem")["results"]) == 1
    assert search.query("Lorem")["results"][0]["id"] == ids1[0]
    assert len(search.query("am*")["results"]) == 1
    assert search.query("am*")["results"][0]["id"] == ids2[0]
    assert len(search.query("Lorem or amet")["results"]) == 2
    # assert search.count() == 2
    search.count()


def test_query_multiple(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["Lorem ipsum dolor"])
    search.add(["sit amet"])
    assert len(search.query("Lorem ipsum")["results"]) == 1
    assert len(search.query("sit amet")["results"]) == 1
    assert len(search.query("Lorem sit")["results"]) == 0


def test_add_name(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="my_name")
    assert search.query("Lorem")["results"] == []
    search.add(["Lorem ipsum dolor"])
    assert len(search.query("Lorem")["results"]) == 1
    search = CollectionSQLite(path, name="123")
    assert len(search.query("Lorem")["results"]) == 0
    search = CollectionSQLite(path, name="my_name")
    assert len(search.query("Lorem")["results"]) == 1


def test_add_id(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    ids = search.add(["x"])
    assert len(ids) == 1
    assert len(ids[0]) == 36  # is UUIDv4
    ids = search.add(["y"], ids=["my_id"])
    assert ids == ["my_id"]
    res = search.query("y")
    assert len(res["results"]) == 1
    res = res["results"]
    assert res[0]["id"] == "my_id"
    # does not raise, but updates
    search.add(["z"], ids=["my_id"])
    res = search.query("y")
    assert len(res["results"]) == 0
    res = search.query("z")
    assert len(res["results"]) == 1


def test_update(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    ids = search.add(["Lorem ipsum"])
    res = search.query("Lorem")
    res = res["results"]
    assert len(res) == 1
    assert res[0]["id"] == ids[0]
    search.update(ids=ids, contents=["dolor sit"])
    res = search.query("Lorem")
    assert len(res["results"]) == 0
    res = search.query("sit")
    res = res["results"]
    assert len(res) == 1
    assert res[0]["id"] == ids[0]


def test_delete(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.count()
    ids = search.add(["Lorem ipsum", "Lorem dolor"])
    res = search.query("Lorem")
    assert len(res["results"]) == 2
    search.count()
    search.delete(ids)
    res = search.query("Lorem")
    assert len(res["results"]) == 0
    search.count()
    search.delete(ids)


def test_query_metadata(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["Lorem ipsum dolor"], metadatas=[{"foo": "bar"}])
    search.add(["sit amet"])
    res = search.query("Lorem")
    res = res["results"]
    assert len(res) == 1
    assert res[0]["metadata"] == {"foo": "bar"}
    res = search.query("sit")
    assert res["total"] == 1
    res = res["results"]
    assert len(res) == 1
    assert res[0]["metadata"] is None


def test_query_order(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["Lorem"], metadatas=[{"k1": "a", "k2": "c"}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": "b", "k2": "c"}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": "c", "k2": "c"}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": "d", "k2": "b"}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": "e", "k2": "b"}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": "f", "k2": "b"}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": "g", "k2": "a"}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": "h", "k2": "a"}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": "i", "k2": "a"}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    res = search.query("Lorem")
    assert res["total"] == 10
    assert len(res["results"]) == 10
    # k1
    res = search.query("Lorem", order_by="k1")
    assert res["total"] == 10
    res = res["results"]
    assert len(res) == 10
    assert [r["id"][1:] for r in res] == list("1234567890")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("abcdefghi0")
    # +k1
    res = search.query("Lorem", order_by="+k1")["results"]
    assert [r["id"][1:] for r in res] == list("1234567890")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("abcdefghi0")
    # -k1
    res = search.query("Lorem", order_by="-k1")["results"]
    assert [r["id"][1:] for r in res] == list("0987654321")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("0ihgfedcba")
    # k2,k1
    res = search.query("Lorem", order_by=["k2", "k1"])["results"]
    assert [r["id"][1:] for r in res] == list("7894561230")
    assert [(r["metadata"] or {}).get("k2", "0") for r in res] == list("aaabbbccc0")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("ghidefabc0")
    # k2,-k1
    res = search.query("Lorem", order_by=["k2", "-k1"])["results"]
    assert [r["id"][1:] for r in res] == list("9876543210")
    assert [(r["metadata"] or {}).get("k2", "0") for r in res] == list("aaabbbccc0")
    assert [(r["metadata"] or {}).get("k1", "0") for r in res] == list("ihgfedcba0")


def test_query_limit_offset(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["Lorem"], metadatas=[{"k1": "a", "k2": "c"}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": "b", "k2": "c"}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": "c", "k2": "c"}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": "d", "k2": "b"}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": "e", "k2": "b"}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": "f", "k2": "b"}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": "g", "k2": "a"}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": "h", "k2": "a"}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": "i", "k2": "a"}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    res = search.query("Lorem", order_by="k1")
    assert res["total"] == 10
    assert len(res["results"]) == 10
    res = search.query("Lorem", order_by="k1", limit=0)
    assert res["total"] == 10
    assert len(res["results"]) == 10
    res = search.query("Lorem", order_by="k1", limit=3)
    assert res["total"] == 10
    res = res["results"]
    assert len(res) == 3
    assert [r["id"][1:] for r in res] == list("123")
    res = search.query("Lorem", order_by="k1", limit=3, offset=3)
    assert res["total"] == 10
    res = res["results"]
    assert len(res) == 3
    assert [r["id"][1:] for r in res] == list("456")
    res = search.query("Lorem", order_by="k1", limit=3, offset=8)
    assert res["total"] == 10
    res = res["results"]
    assert len(res) == 2
    assert [r["id"][1:] for r in res] == list("90")


def test_query_where(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["Lorem"], metadatas=[{"k1": "a", "k2": "c"}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": "b", "k2": "c"}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": "c", "k2": "c"}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": "d", "k2": "b"}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": "e", "k2": "b"}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": "f", "k2": "b"}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": "g", "k2": "a"}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": "h", "k2": "a"}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": "i", "k2": "a"}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    res = search.query("Lorem", where={"k2": "a"}, order_by="k1")
    assert res["total"] == 3
    res = res["results"]
    assert len(res) == 3
    res = search.query("Lorem", where={"k2": {"$eq": "a"}}, order_by="k1")
    assert res["total"] == 3
    res = res["results"]
    assert len(res) == 3
    res = search.query("Lorem", where={"k2": {"$gt": "a"}}, order_by="k1")
    assert res["total"] == 6
    res = res["results"]
    assert len(res) == 6
    res = search.query("Lorem", where={"k2": {"$lt": "a"}}, order_by="k1")
    assert res["total"] == 0
    res = res["results"]
    assert len(res) == 0


def test_query_where_num(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["Lorem"], metadatas=[{"k1": 1, "k2": 3}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": 2, "k2": 3}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": 3, "k2": 3}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": 4, "k2": 2}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": 5, "k2": 2}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": 6, "k2": 2}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": 7, "k2": 1}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": 8, "k2": 1}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": 9, "k2": 1}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    res = search.query("Lorem", where={"k2": 1}, order_by="k1")
    assert res["total"] == 3
    res = res["results"]
    assert len(res) == 3
    res = search.query("Lorem", where={"k2": {"$eq": 1}}, order_by="k1")
    assert res["total"] == 3
    res = res["results"]
    assert len(res) == 3
    res = search.query("Lorem", where={"k2": {"$gt": 1}}, order_by="k1")
    assert res["total"] == 6
    res = res["results"]
    assert len(res) == 6
    res = search.query("Lorem", where={"k2": {"$lt": 1}}, order_by="k1")
    assert res["total"] == 0
    res = res["results"]
    assert len(res) == 0


def test_query_where_in(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["Lorem"], metadatas=[{"k1": "a", "k2": "c"}], ids=["i1"])
    search.add(["Lorem"], metadatas=[{"k1": "b", "k2": "c"}], ids=["i2"])
    search.add(["Lorem"], metadatas=[{"k1": "c", "k2": "c"}], ids=["i3"])
    search.add(["Lorem"], metadatas=[{"k1": "d", "k2": "b"}], ids=["i4"])
    search.add(["Lorem"], metadatas=[{"k1": "e", "k2": "b"}], ids=["i5"])
    search.add(["Lorem"], metadatas=[{"k1": "f", "k2": "b"}], ids=["i6"])
    search.add(["Lorem"], metadatas=[{"k1": "g", "k2": "a"}], ids=["i7"])
    search.add(["Lorem"], metadatas=[{"k1": "h", "k2": "a"}], ids=["i8"])
    search.add(["Lorem"], metadatas=[{"k1": "i", "k2": "a"}], ids=["i9"])
    search.add(["Lorem"], ids=["i0"])
    with pytest.raises(ValueError):
        # wrong operator
        search.query("Lorem", where={"k1": {"in": "a"}})
    res = search.query(
        "Lorem", where={"k1": {"$in": ["a", "b", "c", "d"]}}, order_by="k1"
    )
    assert res["total"] == 4
    res = res["results"]
    assert len(res) == 4
    assert [r["id"][1:] for r in res] == list("1234")
    res = search.query(
        "Lorem", where={"k1": {"$nin": ["a", "b", "c", "d"]}}, order_by="k1"
    )
    assert res["total"] == 5
    res = res["results"]
    assert len(res) == 5
    assert [r["id"][1:] for r in res] == list("56789")


def test_all_docs(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["Lorem ipsum dolor"])
    search.add(["sit amet"])
    res = search.get()
    assert len(res["results"]) == 2
    assert res["total"] == 2


def test_vector_add(tmp_path):
    path = tmp_path / "search_engine.db"
    vectors = {"Lorem ipsum dolor": [0, 0, 0], "sit amet": [0, 0.5, 0]}

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionSQLite(path, name="vector", embedding_function=f)
    search.add(["Lorem ipsum dolor", "sit amet"])
    with search.conn() as conn:
        res = conn.execute("SELECT embedding FROM documents")
        vectors = res.fetchall()
        vectors = [np.frombuffer(v[0], dtype=np.float32) for v in vectors]
        assert len(vectors) == 2
        assert isinstance(vectors[0], np.ndarray)
        assert vectors[0].tolist() == [0, 0, 0]
        assert vectors[1].tolist() == [0, 0.5, 0]


def test_vector_query(tmp_path):
    path = tmp_path / "search_engine.db"
    vectors = {
        "Lorem ipsum dolor": [1, 1, 1],
        "sit amet": [1, -1, 1],
        "consectetur": [-1, -1, 1],
        "adipiscing": [-1, -1, -1],
    }

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionSQLite(path, name="vector", embedding_function=f)
    search.add(["Lorem ipsum dolor", "sit amet"])
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    # limit & offset
    res = search.query("consectetur", vector_search=True, offset=0, limit=1)
    assert res["total"] == 2
    assert len(res["results"]) == 1
    assert res["results"][0]["content"] == "sit amet"
    res = search.query("consectetur", vector_search=True, offset=1, limit=1)
    assert res["total"] == 2
    assert len(res["results"]) == 1
    assert res["results"][0]["content"] == "Lorem ipsum dolor"
    res = search.query("consectetur", vector_search=True, offset=2)
    assert res["total"] == 2
    assert len(res["results"]) == 0


def test_vector_query_fts(tmp_path):
    path = tmp_path / "search_engine.db"
    vectors = {
        "Lorem ipsum dolor": [1, 1, 1],
        "sit amet": [1, -1, 1],
        "consectetur": [-1, -1, 1],
        "adipiscing": [-1, -1, -1],
    }

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionSQLite(path, name="vector", embedding_function=f)
    search.add(["Lorem ipsum dolor", "sit amet"])
    res = search.query("Lorem", vector_search=False)
    assert res["total"] == 1
    assert res["results"][0]["content"] == "Lorem ipsum dolor"


def test_vector_update(tmp_path):
    path = tmp_path / "search_engine.db"
    vectors = {
        "Lorem ipsum dolor": [1, 1, 1],
        "sit amet": [1, -1, 1],
        "consectetur": [-1, -1, 1],
        "adipiscing": [-1, -1, -1],
    }

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionSQLite(path, name="vector", embedding_function=f)
    ids = search.add(["Lorem ipsum dolor", "sit amet"])
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][0]["id"] == ids[1]
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    assert res["results"][1]["id"] == ids[0]
    # update: switch order
    search.update(ids=ids, contents=["sit amet", "Lorem ipsum dolor"])
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][0]["id"] == ids[0]
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    assert res["results"][1]["id"] == ids[1]


def test_vector_update_nofts(tmp_path):
    path = tmp_path / "search_engine.db"
    vectors = {
        "Lorem ipsum dolor": [1, 1, 1],
        "sit amet": [1, -1, 1],
        "consectetur": [-1, -1, 1],
        "adipiscing": [-1, -1, -1],
    }

    def f(documents):
        return [vectors[doc] for doc in documents]

    search = CollectionSQLite(path, name="vector", embedding_function=f, use_fts=False)
    ids = search.add(["Lorem ipsum dolor", "sit amet"])
    with pytest.raises(ValueError):
        res = search.query("Lorem", vector_search=False)
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][0]["id"] == ids[1]
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    assert res["results"][1]["id"] == ids[0]
    # update: switch order
    search.update(ids=ids, contents=["sit amet", "Lorem ipsum dolor"])
    res = search.query("consectetur", vector_search=True)
    assert res["total"] == 2
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["content"] == "sit amet"
    assert res["results"][0]["rank"] == pytest.approx(1 / 3)
    assert res["results"][0]["id"] == ids[0]
    assert res["results"][1]["content"] == "Lorem ipsum dolor"
    assert res["results"][1]["rank"] == pytest.approx(-1 / 3)
    assert res["results"][1]["id"] == ids[1]


def test_query_hyphenated_word(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["This is a test-word example"])
    search.add(["Another document without it"])
    res = search.query("test-word")
    assert res["total"] == 1
    assert len(res["results"]) == 1
    assert res["results"][0]["content"] == "This is a test-word example"


def test_query_hyphenated_wildcard(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["This is a test-word example"])
    search.add(["This is a test-word-extended version"])
    search.add(["Another document without it"])
    res = search.query("test-word*")
    assert res["total"] == 2
    assert len(res["results"]) == 2


def test_query_wildcard_end(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["testing document"])
    search.add(["tester document"])
    search.add(["tests document"])
    search.add(["another document"])
    res = search.query("test*")
    assert res["total"] == 3
    assert len(res["results"]) == 3


def test_query_wildcard_middle(tmp_path):
    # Note: SQLite FTS5 does not support wildcards in the middle of words
    # Only trailing wildcards are supported
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["testing document"])
    search.add(["tester document"])
    search.add(["another document"])
    res = search.query("te*ing")
    # This won't match anything because mid-word wildcards aren't supported
    assert res["total"] == 0


def test_query_wildcard_beginning(tmp_path):
    # Note: SQLite FTS5 does not support leading wildcards
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["testing document"])
    search.add(["interesting document"])
    search.add(["another document"])
    res = search.query("*sting")
    # This won't match anything because leading wildcards aren't supported
    assert res["total"] == 0


def test_query_apostrophe(tmp_path):
    path = tmp_path / "search_engine.db"
    search = CollectionSQLite(path, name="123")
    search.add(["It's a test document"])
    search.add(["Another document"])
    res = search.query("it's")
    assert res["total"] == 1
    assert res["results"][0]["content"] == "It's a test document"



"""Script to assess the performance of the Postgres backend."""

import numpy as np
import subprocess
import time

import psycopg2
from psycopg2 import OperationalError
from sifts import Collection
from profiling_sqlite import run_add_update_delete, timer


def is_postgres_healthy():
    try:
        conn = psycopg2.connect(
            dbname="testdb",
            user="testuser",
            password="testpass",
            host="localhost",
            port=5432,
        )
        conn.close()
        return True
    except OperationalError:
        return False


def run_timing():
    """Run timing."""
    subprocess.run(["docker", "compose", "up", "-d"])

    timeout = 30
    pause = 0.5
    start_time = time.time()

    while time.time() - start_time < timeout:
        if is_postgres_healthy():
            break
        else:
            time.sleep(pause)

    try:

        with timer("Create database table"):
            Collection(
                "postgresql://testuser:testpass@localhost:5432/testdb", name="123"
            )

        with timer("Instantiate again"):
            engine = Collection(
                "postgresql://testuser:testpass@localhost:5432/testdb", name="123"
            )

        print("-- Full-text search --")
        run_add_update_delete(engine)

        def f(documents):
            return [np.random.rand(384) for _ in documents]

        engine = Collection(
            "postgresql://testuser:testpass@localhost:5432/testdb",
            name="456",
            embedding_function=f,
            use_fts=False,
        )

        print("-- Vector search --")
        run_add_update_delete(engine)

        engine = Collection(
            "postgresql://testuser:testpass@localhost:5432/testdb",
            name="456",
            embedding_function=f,
        )

        print("-- Both --")
        run_add_update_delete(engine)

    except:
        subprocess.run(["docker", "compose", "down"])
        raise
    finally:
        subprocess.run(["docker", "compose", "down"])


if __name__ == "__main__":
    run_timing()
